"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createPackageRequester = void 0;
const fs_1 = require("fs");
const path_1 = __importDefault(require("path"));
const store_cafs_1 = require("@pnpm/store.cafs");
const core_loggers_1 = require("@pnpm/core-loggers");
const pick_fetcher_1 = require("@pnpm/pick-fetcher");
const error_1 = require("@pnpm/error");
const graceful_fs_1 = __importDefault(require("@pnpm/graceful-fs"));
const logger_1 = require("@pnpm/logger");
const package_is_installable_1 = require("@pnpm/package-is-installable");
const read_package_json_1 = require("@pnpm/read-package-json");
const dependency_path_1 = require("@pnpm/dependency-path");
const worker_1 = require("@pnpm/worker");
const p_queue_1 = __importDefault(require("p-queue"));
const p_defer_1 = __importDefault(require("p-defer"));
const promise_share_1 = __importDefault(require("promise-share"));
const pick_1 = __importDefault(require("ramda/src/pick"));
const semver_1 = __importDefault(require("semver"));
const ssri_1 = __importDefault(require("ssri"));
const equalOrSemverEqual_1 = require("./equalOrSemverEqual");
const TARBALL_INTEGRITY_FILENAME = 'tarball-integrity';
const packageRequestLogger = (0, logger_1.logger)('package-requester');
const pickBundledManifest = (0, pick_1.default)([
    'bin',
    'bundledDependencies',
    'bundleDependencies',
    'dependencies',
    'directories',
    'engines',
    'name',
    'optionalDependencies',
    'os',
    'peerDependencies',
    'peerDependenciesMeta',
    'scripts',
    'version',
]);
function normalizeBundledManifest(manifest) {
    return {
        ...pickBundledManifest(manifest),
        version: semver_1.default.clean(manifest.version ?? '0.0.0', { loose: true }) ?? manifest.version,
    };
}
function createPackageRequester(opts) {
    opts = opts || {};
    const networkConcurrency = opts.networkConcurrency ?? 16;
    const requestsQueue = new p_queue_1.default({
        concurrency: networkConcurrency,
    });
    const cafsDir = path_1.default.join(opts.storeDir, 'files');
    const getFilePathInCafs = store_cafs_1.getFilePathInCafs.bind(null, cafsDir);
    const fetch = fetcher.bind(null, opts.fetchers, opts.cafs);
    const fetchPackageToStore = fetchToStore.bind(null, {
        readPkgFromCafs: worker_1.readPkgFromCafs.bind(null, cafsDir, opts.verifyStoreIntegrity),
        fetch,
        fetchingLocker: new Map(),
        getFilePathByModeInCafs: store_cafs_1.getFilePathByModeInCafs.bind(null, cafsDir),
        getFilePathInCafs,
        requestsQueue: Object.assign(requestsQueue, {
            counter: 0,
            concurrency: networkConcurrency,
        }),
        relinkLocalDirDeps: opts.relinkLocalDirDeps ?? true,
        storeDir: opts.storeDir,
    });
    const requestPackage = resolveAndFetch.bind(null, {
        engineStrict: opts.engineStrict,
        nodeVersion: opts.nodeVersion,
        pnpmVersion: opts.pnpmVersion,
        force: opts.force,
        fetchPackageToStore,
        requestsQueue,
        resolve: opts.resolve,
        storeDir: opts.storeDir,
    });
    return Object.assign(requestPackage, {
        fetchPackageToStore,
        getFilesIndexFilePath: getFilesIndexFilePath.bind(null, {
            getFilePathInCafs,
            storeDir: opts.storeDir,
        }),
        requestPackage,
    });
}
exports.createPackageRequester = createPackageRequester;
async function resolveAndFetch(ctx, wantedDependency, options) {
    let latest;
    let manifest;
    let normalizedPref;
    let resolution = options.currentPkg?.resolution;
    let pkgId = options.currentPkg?.id;
    const skipResolution = resolution && !options.update;
    let forceFetch = false;
    let updated = false;
    let resolvedVia;
    let publishedAt;
    // When fetching is skipped, resolution cannot be skipped.
    // We need the package's manifest when doing `lockfile-only` installs.
    // When we don't fetch, the only way to get the package's manifest is via resolving it.
    //
    // The resolution step is never skipped for local dependencies.
    if (!skipResolution || options.skipFetch === true || Boolean(pkgId?.startsWith('file:')) || wantedDependency.optional === true) {
        const resolveResult = await ctx.requestsQueue.add(async () => ctx.resolve(wantedDependency, {
            alwaysTryWorkspacePackages: options.alwaysTryWorkspacePackages,
            defaultTag: options.defaultTag,
            publishedBy: options.publishedBy,
            pickLowestVersion: options.pickLowestVersion,
            lockfileDir: options.lockfileDir,
            preferredVersions: options.preferredVersions,
            preferWorkspacePackages: options.preferWorkspacePackages,
            projectDir: options.projectDir,
            registry: options.registry,
            workspacePackages: options.workspacePackages,
        }), { priority: options.downloadPriority });
        manifest = resolveResult.manifest;
        latest = resolveResult.latest;
        resolvedVia = resolveResult.resolvedVia;
        publishedAt = resolveResult.publishedAt;
        // If the integrity of a local tarball dependency has changed,
        // the local tarball should be unpacked, so a fetch to the store should be forced
        forceFetch = Boolean(((options.currentPkg?.resolution) != null) &&
            pkgId?.startsWith('file:') &&
            (options.currentPkg?.resolution).integrity !== resolveResult.resolution.integrity);
        updated = pkgId !== resolveResult.id || !resolution || forceFetch;
        resolution = resolveResult.resolution;
        pkgId = resolveResult.id;
        normalizedPref = resolveResult.normalizedPref;
    }
    const id = pkgId;
    if (resolution.type === 'directory' && !id.startsWith('file:')) {
        if (manifest == null) {
            throw new Error(`Couldn't read package.json of local dependency ${wantedDependency.alias ? wantedDependency.alias + '@' : ''}${wantedDependency.pref ?? ''}`);
        }
        return {
            body: {
                id,
                isLocal: true,
                manifest,
                normalizedPref,
                resolution: resolution,
                resolvedVia,
                updated,
            },
        };
    }
    const isInstallable = (ctx.force === true ||
        (manifest == null
            ? undefined
            : (0, package_is_installable_1.packageIsInstallable)(id, manifest, {
                engineStrict: ctx.engineStrict,
                lockfileDir: options.lockfileDir,
                nodeVersion: ctx.nodeVersion,
                optional: wantedDependency.optional === true,
                pnpmVersion: ctx.pnpmVersion,
            })));
    // We can skip fetching the package only if the manifest
    // is present after resolution
    if ((options.skipFetch === true || isInstallable === false) && (manifest != null)) {
        return {
            body: {
                id,
                isLocal: false,
                isInstallable: isInstallable ?? undefined,
                latest,
                manifest,
                normalizedPref,
                resolution,
                resolvedVia,
                updated,
                publishedAt,
            },
        };
    }
    const pkg = (0, pick_1.default)(['name', 'version'], manifest ?? {});
    const fetchResult = ctx.fetchPackageToStore({
        fetchRawManifest: true,
        force: forceFetch,
        ignoreScripts: options.ignoreScripts,
        lockfileDir: options.lockfileDir,
        pkg: {
            ...pkg,
            id,
            resolution,
        },
        expectedPkg: options.expectedPkg?.name != null
            ? (updated ? { name: options.expectedPkg.name, version: pkg.version } : options.expectedPkg)
            : pkg,
    });
    return {
        body: {
            id,
            isLocal: false,
            isInstallable: isInstallable ?? undefined,
            latest,
            manifest,
            normalizedPref,
            resolution,
            resolvedVia,
            updated,
            publishedAt,
        },
        fetching: fetchResult.fetching,
        filesIndexFile: fetchResult.filesIndexFile,
    };
}
function getFilesIndexFilePath(ctx, opts) {
    const targetRelative = (0, dependency_path_1.depPathToFilename)(opts.pkg.id);
    const target = path_1.default.join(ctx.storeDir, targetRelative);
    const filesIndexFile = opts.pkg.resolution.integrity
        ? ctx.getFilePathInCafs(opts.pkg.resolution.integrity, 'index')
        : path_1.default.join(target, opts.ignoreScripts ? 'integrity-not-built.json' : 'integrity.json');
    return { filesIndexFile, target };
}
function fetchToStore(ctx, opts) {
    if (!opts.pkg.name) {
        opts.fetchRawManifest = true;
    }
    if (!ctx.fetchingLocker.has(opts.pkg.id)) {
        const fetching = (0, p_defer_1.default)();
        const { filesIndexFile, target } = getFilesIndexFilePath(ctx, opts);
        doFetchToStore(filesIndexFile, fetching, target); // eslint-disable-line
        ctx.fetchingLocker.set(opts.pkg.id, {
            fetching: removeKeyOnFail(fetching.promise),
            filesIndexFile,
            fetchRawManifest: opts.fetchRawManifest,
        });
        // When files resolves, the cached result has to set fromStore to true, without
        // affecting previous invocations: so we need to replace the cache.
        //
        // Changing the value of fromStore is needed for correct reporting of `pnpm server`.
        // Otherwise, if a package was not in store when the server started, it will always be
        // reported as "downloaded" instead of "reused".
        fetching.promise.then((cache) => {
            core_loggers_1.progressLogger.debug({
                packageId: opts.pkg.id,
                requester: opts.lockfileDir,
                status: cache.files.fromStore
                    ? 'found_in_store'
                    : 'fetched',
            });
            // If it's already in the store, we don't need to update the cache
            if (cache.files.fromStore) {
                return;
            }
            const tmp = ctx.fetchingLocker.get(opts.pkg.id);
            // If fetching failed then it was removed from the cache.
            // It is OK. In that case there is no need to update it.
            if (tmp == null)
                return;
            ctx.fetchingLocker.set(opts.pkg.id, {
                ...tmp,
                fetching: Promise.resolve({
                    ...cache,
                    files: {
                        ...cache.files,
                        fromStore: true,
                    },
                }),
            });
        })
            .catch(() => {
            ctx.fetchingLocker.delete(opts.pkg.id);
        });
    }
    const result = ctx.fetchingLocker.get(opts.pkg.id);
    if (opts.fetchRawManifest && !result.fetchRawManifest) {
        result.fetching = removeKeyOnFail(result.fetching.then(async ({ files }) => {
            if (!files.filesIndex['package.json'])
                return {
                    files,
                    bundledManifest: undefined,
                };
            if (files.unprocessed) {
                const { integrity, mode } = files.filesIndex['package.json'];
                const manifestPath = ctx.getFilePathByModeInCafs(integrity, mode);
                return {
                    files,
                    bundledManifest: await readBundledManifest(manifestPath),
                };
            }
            return {
                files,
                bundledManifest: await readBundledManifest(files.filesIndex['package.json']),
            };
        }));
        result.fetchRawManifest = true;
    }
    return {
        fetching: (0, promise_share_1.default)(result.fetching),
        filesIndexFile: result.filesIndexFile,
    };
    async function removeKeyOnFail(p) {
        try {
            return await p;
        }
        catch (err) { // eslint-disable-line
            ctx.fetchingLocker.delete(opts.pkg.id);
            throw err;
        }
    }
    async function doFetchToStore(filesIndexFile, fetching, target) {
        try {
            const isLocalTarballDep = opts.pkg.id.startsWith('file:');
            const isLocalPkg = opts.pkg.resolution.type === 'directory';
            if (!opts.force &&
                (!isLocalTarballDep ||
                    await tarballIsUpToDate(opts.pkg.resolution, target, opts.lockfileDir) // eslint-disable-line
                ) &&
                !isLocalPkg) {
                const { verified, pkgFilesIndex, manifest } = await ctx.readPkgFromCafs(filesIndexFile, opts.fetchRawManifest);
                if (verified) {
                    if ((pkgFilesIndex.name != null &&
                        opts.expectedPkg?.name != null &&
                        pkgFilesIndex.name.toLowerCase() !== opts.expectedPkg.name.toLowerCase()) ||
                        (pkgFilesIndex.version != null &&
                            opts.expectedPkg?.version != null &&
                            // We used to not normalize the package versions before writing them to the lockfile and store.
                            // So it may happen that the version will be in different formats.
                            // For instance, v1.0.0 and 1.0.0
                            // Hence, we need to use semver.eq() to compare them.
                            !(0, equalOrSemverEqual_1.equalOrSemverEqual)(pkgFilesIndex.version, opts.expectedPkg.version))) {
                        throw new error_1.PnpmError('UNEXPECTED_PKG_CONTENT_IN_STORE', `\
Package name mismatch found while reading ${JSON.stringify(opts.pkg.resolution)} from the store. \
This means that the lockfile is broken. Expected package: ${opts.expectedPkg.name}@${opts.expectedPkg.version}. \
Actual package in the store by the given integrity: ${pkgFilesIndex.name}@${pkgFilesIndex.version}.`);
                    }
                    fetching.resolve({
                        files: {
                            unprocessed: true,
                            filesIndex: pkgFilesIndex.files,
                            fromStore: true,
                            sideEffects: pkgFilesIndex.sideEffects,
                        },
                        bundledManifest: manifest == null ? manifest : normalizeBundledManifest(manifest),
                    });
                    return;
                }
                if ((pkgFilesIndex?.files) != null) {
                    packageRequestLogger.warn({
                        message: `Refetching ${target} to store. It was either modified or had no integrity checksums`,
                        prefix: opts.lockfileDir,
                    });
                }
            }
            // We fetch into targetStage directory first and then fs.rename() it to the
            // target directory.
            // Tarballs are requested first because they are bigger than metadata files.
            // However, when one line is left available, allow it to be picked up by a metadata request.
            // This is done in order to avoid situations when tarballs are downloaded in chunks
            // As many tarballs should be downloaded simultaneously as possible.
            const priority = (++ctx.requestsQueue.counter % ctx.requestsQueue.concurrency === 0 ? -1 : 1) * 1000;
            const fetchedPackage = await ctx.requestsQueue.add(async () => ctx.fetch(opts.pkg.id, opts.pkg.resolution, {
                filesIndexFile,
                lockfileDir: opts.lockfileDir,
                readManifest: opts.fetchRawManifest,
                onProgress: (downloaded) => {
                    core_loggers_1.fetchingProgressLogger.debug({
                        downloaded,
                        packageId: opts.pkg.id,
                        status: 'in_progress',
                    });
                },
                onStart: (size, attempt) => {
                    core_loggers_1.fetchingProgressLogger.debug({
                        attempt,
                        packageId: opts.pkg.id,
                        size,
                        status: 'started',
                    });
                },
                pkg: {
                    name: opts.pkg.name,
                    version: opts.pkg.version,
                },
            }), { priority });
            if (isLocalTarballDep && opts.pkg.resolution.integrity) {
                await fs_1.promises.mkdir(target, { recursive: true });
                await graceful_fs_1.default.writeFile(path_1.default.join(target, TARBALL_INTEGRITY_FILENAME), opts.pkg.resolution.integrity, 'utf8');
            }
            fetching.resolve({
                files: {
                    local: fetchedPackage.local,
                    fromStore: !fetchedPackage.local ? false : !ctx.relinkLocalDirDeps,
                    filesIndex: fetchedPackage.filesIndex,
                    packageImportMethod: fetchedPackage.packageImportMethod,
                },
                bundledManifest: fetchedPackage.manifest == null ? fetchedPackage.manifest : normalizeBundledManifest(fetchedPackage.manifest),
            });
        }
        catch (err) { // eslint-disable-line
            fetching.reject(err);
        }
    }
}
async function readBundledManifest(pkgJsonPath) {
    return pickBundledManifest(await (0, read_package_json_1.readPackageJson)(pkgJsonPath));
}
async function tarballIsUpToDate(resolution, pkgInStoreLocation, lockfileDir) {
    let currentIntegrity;
    try {
        currentIntegrity = (await graceful_fs_1.default.readFile(path_1.default.join(pkgInStoreLocation, TARBALL_INTEGRITY_FILENAME), 'utf8'));
    }
    catch (err) { // eslint-disable-line
        return false;
    }
    if (resolution.integrity && currentIntegrity !== resolution.integrity)
        return false;
    const tarball = path_1.default.join(lockfileDir, resolution.tarball.slice(5));
    const tarballStream = (0, fs_1.createReadStream)(tarball);
    try {
        return Boolean(await ssri_1.default.checkStream(tarballStream, currentIntegrity));
    }
    catch (err) { // eslint-disable-line
        return false;
    }
}
async function fetcher(fetcherByHostingType, cafs, packageId, resolution, opts) {
    const fetch = (0, pick_fetcher_1.pickFetcher)(fetcherByHostingType, resolution);
    try {
        return await fetch(cafs, resolution, opts); // eslint-disable-line @typescript-eslint/no-explicit-any
    }
    catch (err) { // eslint-disable-line
        packageRequestLogger.warn({
            message: `Fetching ${packageId} failed!`,
            prefix: opts.lockfileDir,
        });
        throw err;
    }
}
//# sourceMappingURL=packageRequester.js.map